@@ -41,8 +41,6 @@
 #if DEBUG_RENDER
 #undef DBG
 #define DBG(x) ErrorF x
-#else
-#define NDEBUG 1
 #endif
 
 #define NO_COMPOSITE 0
@@ -1203,7 +1201,8 @@ gen2_composite_set_target(struct sna_composite_op *op,
 		return FALSE;
 
 	op->dst.bo = priv->gpu_bo;
-	if (!priv->gpu_only)
+	if (!sna_damage_is_all(&priv->gpu_damage,
+			       op->dst.width, op->dst.height))
 		op->damage = &priv->gpu_damage;
 
 	get_drawable_deltas(dst->pDrawable, op->dst.pixmap,
@@ -1253,6 +1252,12 @@ gen2_render_composite(struct sna *sna,
 {
 	DBG(("%s()\n", __FUNCTION__));
 
+	if (op >= ARRAY_SIZE(gen2_blend_op)) {
+		DBG(("%s: fallback due to unhandled blend op: %d\n",
+		     __FUNCTION__, op));
+		return FALSE;
+	}
+
 	/* Try to use the BLT engine unless it implies a
 	 * 3D -> 2D context switch.
 	 */
@@ -1266,12 +1271,6 @@ gen2_render_composite(struct sna *sna,
 			      tmp))
 		return TRUE;
 
-	if (op >= ARRAY_SIZE(gen2_blend_op)) {
-		DBG(("%s: fallback due to unhandled blend op: %d\n",
-		     __FUNCTION__, op));
-		return FALSE;
-	}
-
 	if (!gen2_check_dst_format(dst->format)) {
 		DBG(("%s: fallback due to unhandled dst format: %x\n",
 		     __FUNCTION__, dst->format));
@@ -1291,6 +1290,7 @@ gen2_render_composite(struct sna *sna,
 		     __FUNCTION__));
 		return FALSE;
 	}
+	sna_render_reduce_damage(tmp, dst_x, dst_y, width, height);
 
 	tmp->op = op;
 	if (tmp->dst.width > 2048 ||
@@ -1686,6 +1686,7 @@ gen2_render_composite_spans(struct sna *sna,
 			    int16_t src_x,  int16_t src_y,
 			    int16_t dst_x,  int16_t dst_y,
 			    int16_t width,  int16_t height,
+			    unsigned flags,
 			    struct sna_composite_spans_op *tmp)
 {
 	DBG(("%s(src=(%d, %d), dst=(%d, %d), size=(%d, %d))\n", __FUNCTION__,
@@ -1715,6 +1716,7 @@ gen2_render_composite_spans(struct sna *sna,
 		     __FUNCTION__));
 		return FALSE;
 	}
+	sna_render_reduce_damage(&tmp->base, dst_x, dst_y, width, height);
 
 	tmp->base.op = op;
 	if (tmp->base.dst.width > 2048 ||
@@ -1907,6 +1909,12 @@ gen2_render_fill_boxes(struct sna *sna,
 	struct sna_composite_op tmp;
 	uint32_t pixel;
 
+	if (op >= ARRAY_SIZE(gen2_blend_op)) {
+		DBG(("%s: fallback due to unhandled blend op: %d\n",
+		     __FUNCTION__, op));
+		return FALSE;
+	}
+
 #if NO_FILL_BOXES
 	return gen2_render_fill_boxes_try_blt(sna, op, format, color,
 					      dst, dst_bo,
@@ -1917,12 +1925,6 @@ gen2_render_fill_boxes(struct sna *sna,
 	     __FUNCTION__, op, (int)format,
 	     color->red, color->green, color->blue, color->alpha));
 
-	if (op >= ARRAY_SIZE(gen2_blend_op)) {
-		DBG(("%s: fallback due to unhandled blend op: %d\n",
-		     __FUNCTION__, op));
-		return FALSE;
-	}
-
 	if (dst->drawable.width > 2048 ||
 	    dst->drawable.height > 2048 ||
 	    dst_bo->pitch > 8192 ||
