@@ -72,8 +72,6 @@ static inline void list_replace(struct list *old,
 #if DEBUG_KGEM
 #undef DBG
 #define DBG(x) ErrorF x
-#else
-#define NDEBUG 1
 #endif
 
 #define PAGE_SIZE 4096
@@ -113,6 +111,7 @@ static int gem_set_tiling(int fd, uint32_t handle, int tiling, int stride)
 	if (DBG_NO_TILING)
 		return I915_TILING_NONE;
 
+	VG_CLEAR(set_tiling);
 	do {
 		set_tiling.handle = handle;
 		set_tiling.tiling_mode = tiling;
@@ -132,6 +131,7 @@ static void *gem_mmap(int fd, uint32_t handle, int size, int prot)
 	DBG(("%s(handle=%d, size=%d, prot=%s)\n", __FUNCTION__,
 	     handle, size, prot & PROT_WRITE ? "read/write" : "read-only"));
 
+	VG_CLEAR(mmap_arg);
 	mmap_arg.handle = handle;
 	if (drmIoctl(fd, DRM_IOCTL_I915_GEM_MMAP_GTT, &mmap_arg)) {
 		assert(0);
@@ -144,6 +144,7 @@ static void *gem_mmap(int fd, uint32_t handle, int size, int prot)
 		ptr = NULL;
 	}
 
+	VG_CLEAR(set_domain);
 	set_domain.handle = handle;
 	set_domain.read_domains = I915_GEM_DOMAIN_GTT;
 	set_domain.write_domain = prot & PROT_WRITE ? I915_GEM_DOMAIN_GTT : 0;
@@ -161,6 +162,7 @@ static int gem_write(int fd, uint32_t handle,
 	DBG(("%s(handle=%d, offset=%d, len=%d)\n", __FUNCTION__,
 	     handle, offset, length));
 
+	VG_CLEAR(pwrite);
 	pwrite.handle = handle;
 	pwrite.offset = offset;
 	pwrite.size = length;
@@ -168,15 +170,17 @@ static int gem_write(int fd, uint32_t handle,
 	return drmIoctl(fd, DRM_IOCTL_I915_GEM_PWRITE, &pwrite);
 }
 
-static int gem_read(int fd, uint32_t handle, const void *dst, int length)
+static int gem_read(int fd, uint32_t handle, const void *dst,
+		    int offset, int length)
 {
 	struct drm_i915_gem_pread pread;
 
 	DBG(("%s(handle=%d, len=%d)\n", __FUNCTION__,
 	     handle, length));
 
+	VG_CLEAR(pread);
 	pread.handle = handle;
-	pread.offset = 0;
+	pread.offset = offset;
 	pread.size = length;
 	pread.data_ptr = (uintptr_t)dst;
 	return drmIoctl(fd, DRM_IOCTL_I915_GEM_PREAD, &pread);
@@ -187,6 +191,7 @@ kgem_busy(struct kgem *kgem, int handle)
 {
 	struct drm_i915_gem_busy busy;
 
+	VG_CLEAR(busy);
 	busy.handle = handle;
 	busy.busy = !kgem->wedged;
 	(void)drmIoctl(kgem->fd, DRM_IOCTL_I915_GEM_BUSY, &busy);
@@ -216,6 +221,7 @@ static uint32_t gem_create(int fd, int size)
 	assert((size & (PAGE_SIZE-1)) == 0);
 #endif
 
+	VG_CLEAR(create);
 	create.handle = 0;
 	create.size = size;
 	(void)drmIoctl(fd, DRM_IOCTL_I915_GEM_CREATE, &create);
@@ -229,6 +235,7 @@ gem_madvise(int fd, uint32_t handle, uint32_t state)
 	struct drm_i915_gem_madvise madv;
 	int ret;
 
+	VG_CLEAR(madv);
 	madv.handle = handle;
 	madv.madv = state;
 	madv.retained = 1;
@@ -242,6 +249,7 @@ static void gem_close(int fd, uint32_t handle)
 {
 	struct drm_gem_close close;
 
+	VG_CLEAR(close);
 	close.handle = handle;
 	(void)drmIoctl(fd, DRM_IOCTL_GEM_CLOSE, &close);
 }
@@ -255,6 +263,7 @@ static struct kgem_bo *__kgem_bo_init(struct kgem_bo *bo,
 	bo->handle = handle;
 	bo->size = size;
 	bo->reusable = true;
+	bo->purgeable = true;
 	bo->cpu_read = true;
 	bo->cpu_write = true;
 	list_init(&bo->request);
@@ -348,11 +357,12 @@ void kgem_init(struct kgem *kgem, int fd, struct pci_device *dev, int gen)
 	if (!DBG_NO_VMAP) {
 		drm_i915_getparam_t gp;
 
+		v = 0;
+		VG_CLEAR(gp);
 		gp.param = I915_PARAM_HAS_VMAP;
 		gp.value = &v;
-		kgem->has_vmap =
-			drmIoctl(kgem->fd, DRM_IOCTL_I915_GETPARAM, &gp) == 0 &&
-			v > 0;
+		drmIoctl(kgem->fd, DRM_IOCTL_I915_GETPARAM, &gp);
+		kgem->has_vmap = v > 0;
 	}
 #endif
 	DBG(("%s: using vmap=%d\n", __FUNCTION__, kgem->has_vmap));
@@ -361,16 +371,19 @@ void kgem_init(struct kgem *kgem, int fd, struct pci_device *dev, int gen)
 		if (!DBG_NO_RELAXED_FENCING) {
 			drm_i915_getparam_t gp;
 
+			v = 0;
+			VG_CLEAR(gp);
 			gp.param = I915_PARAM_HAS_RELAXED_FENCING;
 			gp.value = &v;
-			if (drmIoctl(kgem->fd, DRM_IOCTL_I915_GETPARAM, &gp) == 0)
-				kgem->has_relaxed_fencing = v > 0;
+			drmIoctl(kgem->fd, DRM_IOCTL_I915_GETPARAM, &gp);
+			kgem->has_relaxed_fencing = v > 0;
 		}
 	} else
 		kgem->has_relaxed_fencing = 1;
 	DBG(("%s: has relaxed fencing=%d\n", __FUNCTION__,
 	     kgem->has_relaxed_fencing));
 
+	VG_CLEAR(aperture);
 	aperture.aper_size = 64*1024*1024;
 	(void)drmIoctl(fd, DRM_IOCTL_I915_GEM_GET_APERTURE, &aperture);
 
@@ -393,6 +406,7 @@ void kgem_init(struct kgem *kgem, int fd, struct pci_device *dev, int gen)
 	DBG(("%s: max object size %d\n", __FUNCTION__, kgem->max_object_size));
 
 	v = 8;
+	VG_CLEAR(gp);
 	gp.param = I915_PARAM_NUM_FENCES_AVAIL;
 	gp.value = &v;
 	(void)drmIoctl(fd, DRM_IOCTL_I915_GETPARAM, &gp);
@@ -412,6 +426,7 @@ static uint32_t kgem_get_unique_id(struct kgem *kgem)
 }
 
 static uint32_t kgem_surface_size(struct kgem *kgem,
+				  bool relaxed_fencing,
 				  uint32_t width,
 				  uint32_t height,
 				  uint32_t bpp,
@@ -445,17 +460,32 @@ static uint32_t kgem_surface_size(struct kgem *kgem,
 		break;
 	}
 
+	/* If it is too wide for the blitter, don't even bother.  */
 	*pitch = ALIGN(width * bpp / 8, tile_width);
-	if (kgem->gen < 40 && tiling != I915_TILING_NONE) {
-		if (*pitch > 8192)
+	if (kgem->gen < 40) {
+		if(tiling != I915_TILING_NONE) {
+			if (*pitch > 8192)
+				return 0;
+			for (size = tile_width; size < *pitch; size <<= 1)
+				;
+			*pitch = size;
+		} else {
+			if (*pitch >= 32768)
+				return 0;
+		}
+	} else {
+		int limit = 32768;
+		if (tiling)
+			limit *= 4;
+		if (*pitch >= limit)
 			return 0;
-		for (size = tile_width; size < *pitch; size <<= 1)
-			;
-		*pitch = size;
 	}
+	height = ALIGN(height, tile_height);
+	if (height >= 65536)
+		return 0;
 
-	size = *pitch * ALIGN(height, tile_height);
-	if (kgem->has_relaxed_fencing || tiling == I915_TILING_NONE)
+	size = *pitch * height;
+	if (relaxed_fencing || tiling == I915_TILING_NONE || kgem->gen >= 40)
 		return ALIGN(size, PAGE_SIZE);
 
 	/*  We need to allocate a pot fence region for a tiled buffer. */
@@ -511,8 +541,12 @@ void _kgem_add_bo(struct kgem *kgem, struct kgem_bo *bo)
 	bo->exec = kgem_add_handle(kgem, bo);
 	bo->rq = kgem->next_request;
 	bo->gpu = true;
+
 	list_move(&bo->request, &kgem->next_request->buffers);
+
+	/* XXX is it worth working around gcc here? */
 	kgem->flush |= bo->flush;
+	kgem->sync |= bo->sync;
 }
 
 static uint32_t kgem_end_batch(struct kgem *kgem)
@@ -539,12 +573,29 @@ static void kgem_fixup_self_relocs(struct kgem *kgem, struct kgem_bo *bo)
 	}
 }
 
+static void kgem_bo_free(struct kgem *kgem, struct kgem_bo *bo)
+{
+	struct kgem_bo_binding *b;
+
+	b = bo->binding.next;
+	while (b) {
+		struct kgem_bo_binding *next = b->next;
+		free (b);
+		b = next;
+	}
+
+	list_del(&bo->list);
+	list_del(&bo->request);
+	gem_close(kgem->fd, bo->handle);
+	free(bo);
+}
+
 static void __kgem_bo_destroy(struct kgem *kgem, struct kgem_bo *bo)
 {
 	assert(list_is_empty(&bo->list));
 	assert(bo->refcnt == 0);
 
-	bo->src_bound = bo->dst_bound = 0;
+	bo->binding.offset = 0;
 
 	if (NO_CACHE)
 		goto destroy;
@@ -552,34 +603,33 @@ static void __kgem_bo_destroy(struct kgem *kgem, struct kgem_bo *bo)
 	if(!bo->reusable)
 		goto destroy;
 
-	if (!bo->deleted && !bo->exec) {
+	if (bo->purgeable && !bo->rq && !bo->needs_flush) {
+		assert(!bo->purged);
+
 		if (!gem_madvise(kgem->fd, bo->handle, I915_MADV_DONTNEED)) {
 			kgem->need_purge |= bo->gpu;
 			goto destroy;
 		}
 
-		bo->deleted = 1;
+		bo->purged = true;
 	}
 
 	kgem->need_expire = true;
 	if (bo->rq) {
 		list_move(&bo->list, active(kgem, bo->size));
-	} else if (bo->needs_flush | bo->gpu) {
+	} else if (bo->purged) {
+		list_move(&bo->list, inactive(kgem, bo->size));
+	} else {
 		assert(list_is_empty(&bo->request));
 		list_add(&bo->request, &kgem->flushing);
 		list_move(&bo->list, active(kgem, bo->size));
-	} else {
-		list_move(&bo->list, inactive(kgem, bo->size));
 	}
 
 	return;
 
 destroy:
-	if (!bo->exec) {
-		list_del(&bo->request);
-		gem_close(kgem->fd, bo->handle);
-		free(bo);
-	}
+	if (!bo->exec)
+		kgem_bo_free(kgem, bo);
 }
 
 static void kgem_bo_unref(struct kgem *kgem, struct kgem_bo *bo)
@@ -588,9 +638,10 @@ static void kgem_bo_unref(struct kgem *kgem, struct kgem_bo *bo)
 		__kgem_bo_destroy(kgem, bo);
 }
 
-void kgem_retire(struct kgem *kgem)
+bool kgem_retire(struct kgem *kgem)
 {
 	struct kgem_bo *bo, *next;
+	bool retired = false;
 
 	DBG(("%s\n", __FUNCTION__));
 
@@ -600,10 +651,17 @@ void kgem_retire(struct kgem *kgem)
 
 		DBG(("%s: moving %d from flush to inactive\n",
 		     __FUNCTION__, bo->handle));
-		bo->needs_flush = 0;
-		bo->gpu = false;
-		list_move(&bo->list, inactive(kgem, bo->size));
-		list_del(&bo->request);
+		if (bo->purgeable &&
+		    gem_madvise(kgem->fd, bo->handle, I915_MADV_DONTNEED)) {
+			bo->purged = true;
+			bo->needs_flush = false;
+			bo->gpu = false;
+			list_move(&bo->list, inactive(kgem, bo->size));
+			list_del(&bo->request);
+		} else
+			kgem_bo_free(kgem, bo);
+
+		retired = true;
 	}
 
 	while (!list_is_empty(&kgem->requests)) {
@@ -633,21 +691,29 @@ void kgem_retire(struct kgem *kgem)
 			bo->gpu = bo->needs_flush;
 
 			if (bo->refcnt == 0) {
-				assert(bo->deleted);
-				if (bo->needs_flush) {
-					DBG(("%s: moving %d to flushing\n",
-					     __FUNCTION__, bo->handle));
-					list_add(&bo->request, &kgem->flushing);
-				} else if (bo->reusable) {
-					DBG(("%s: moving %d to inactive\n",
-					     __FUNCTION__, bo->handle));
-					list_move(&bo->list,
-						  inactive(kgem, bo->size));
+				if (bo->reusable) {
+					if (bo->needs_flush || !bo->purgeable) {
+						DBG(("%s: moving %d to flushing\n",
+						     __FUNCTION__, bo->handle));
+						list_add(&bo->request, &kgem->flushing);
+					} else if(gem_madvise(kgem->fd,
+							      bo->handle,
+							      I915_MADV_DONTNEED)) {
+						DBG(("%s: moving %d to inactive\n",
+						     __FUNCTION__, bo->handle));
+						bo->purged = true;
+						list_move(&bo->list,
+							  inactive(kgem, bo->size));
+						retired = true;
+					} else {
+						DBG(("%s: closing %d\n",
+						     __FUNCTION__, bo->handle));
+						kgem_bo_free(kgem, bo);
+					}
 				} else {
 					DBG(("%s: closing %d\n",
 					     __FUNCTION__, bo->handle));
-					gem_close(kgem->fd, bo->handle);
-					free(bo);
+					kgem_bo_free(kgem, bo);
 				}
 			}
 		}
@@ -655,10 +721,10 @@ void kgem_retire(struct kgem *kgem)
 		rq->bo->refcnt--;
 		assert(rq->bo->refcnt == 0);
 		if (gem_madvise(kgem->fd, rq->bo->handle, I915_MADV_DONTNEED)) {
-			rq->bo->deleted = 1;
+			rq->bo->purged = true;
 			assert(rq->bo->gpu == 0);
-			list_move(&rq->bo->list,
-				  inactive(kgem, rq->bo->size));
+			list_move(&rq->bo->list, inactive(kgem, rq->bo->size));
+			retired = true;
 		} else {
 			kgem->need_purge = 1;
 			gem_close(kgem->fd, rq->bo->handle);
@@ -669,8 +735,12 @@ void kgem_retire(struct kgem *kgem)
 		free(rq);
 	}
 
-	if (kgem->ring && list_is_empty(&kgem->requests))
+	kgem->need_retire = !list_is_empty(&kgem->requests);
+	if (!kgem->need_retire && kgem->ring)
 		kgem->ring = kgem->mode;
+	kgem->busy &= kgem->need_retire;
+
+	return retired;
 }
 
 static void kgem_commit(struct kgem *kgem)
@@ -679,48 +749,28 @@ static void kgem_commit(struct kgem *kgem)
 	struct kgem_bo *bo, *next;
 
 	list_for_each_entry_safe(bo, next, &rq->buffers, request) {
-		bo->src_bound = bo->dst_bound = 0;
 		bo->presumed_offset = bo->exec->offset;
+		bo->binding.offset = 0;
 		bo->exec = NULL;
 		bo->dirty = false;
 		bo->cpu_read = false;
 		bo->cpu_write = false;
 
-		if (!bo->refcnt) {
-			if (!bo->reusable) {
-destroy:
-				list_del(&bo->list);
-				list_del(&bo->request);
-				gem_close(kgem->fd, bo->handle);
-				free(bo);
-				continue;
-			}
-			if (!bo->deleted) {
-				if (!gem_madvise(kgem->fd, bo->handle,
-						 I915_MADV_DONTNEED)) {
-					kgem->need_purge = 1;
-					goto destroy;
-				}
-				bo->deleted = 1;
-			}
+		if (!bo->refcnt && !bo->reusable) {
+			kgem_bo_free(kgem, bo);
+			continue;
 		}
 	}
 
 	list_add_tail(&rq->list, &kgem->requests);
 	kgem->next_request = __kgem_request_alloc();
+	kgem->need_retire = 1;
 }
 
 static void kgem_close_list(struct kgem *kgem, struct list *head)
 {
-	while (!list_is_empty(head)) {
-		struct kgem_bo *bo;
-
-		bo = list_first_entry(head, struct kgem_bo, list);
-		gem_close(kgem->fd, bo->handle);
-		list_del(&bo->list);
-		list_del(&bo->request);
-		free(bo);
-	}
+	while (!list_is_empty(head))
+		kgem_bo_free(kgem, list_first_entry(head, struct kgem_bo, list));
 }
 
 static void kgem_close_inactive(struct kgem *kgem)
@@ -741,16 +791,14 @@ static void kgem_finish_partials(struct kgem *kgem)
 				DBG(("%s: discarding unused partial array: %d/%d\n",
 				     __FUNCTION__, bo->used, bo->alloc));
 
-				list_del(&bo->base.list);
-				gem_close(kgem->fd, bo->base.handle);
-				free(bo);
+				kgem_bo_free(kgem, &bo->base);
 			}
 
 			continue;
 		}
 
 		list_del(&bo->base.list);
-		if (bo->write && bo->need_io) {
+		if (bo->need_io) {
 			DBG(("%s: handle=%d, uploading %d/%d\n",
 			     __FUNCTION__, bo->base.handle, bo->used, bo->alloc));
 			assert(!kgem_busy(kgem, bo->base.handle));
@@ -760,11 +808,12 @@ static void kgem_finish_partials(struct kgem *kgem)
 		}
 
 		/* transfer the handle to a minimum bo */
-		if (bo->base.refcnt == 1) {
+		if (bo->base.refcnt == 1 && !bo->base.vmap) {
 			struct kgem_bo *base = malloc(sizeof(*base));
 			if (base) {
 				memcpy(base, &bo->base, sizeof (*base));
 				base->reusable = true;
+				base->purgeable = true;
 				list_init(&base->list);
 				list_replace(&bo->base.request, &base->request);
 				free(bo);
@@ -802,11 +851,8 @@ static void kgem_cleanup(struct kgem *kgem)
 			list_del(&bo->request);
 			bo->rq = NULL;
 			bo->gpu = false;
-			if (bo->refcnt == 0) {
-				list_del(&bo->list);
-				gem_close(kgem->fd, bo->handle);
-				free(bo);
-			}
+			if (bo->refcnt == 0)
+				kgem_bo_free(kgem, bo);
 		}
 
 		list_del(&rq->list);
@@ -855,7 +901,7 @@ void kgem_reset(struct kgem *kgem)
 	while (!list_is_empty(&rq->buffers)) {
 		bo = list_first_entry(&rq->buffers, struct kgem_bo, request);
 
-		bo->src_bound = bo->dst_bound = 0;
+		bo->binding.offset = 0;
 		bo->exec = NULL;
 		bo->dirty = false;
 		bo->cpu_read = false;
@@ -936,6 +982,7 @@ void _kgem_submit(struct kgem *kgem)
 			struct drm_i915_gem_execbuffer2 execbuf;
 			int ret;
 
+			VG_CLEAR(execbuf);
 			execbuf.buffers_ptr = (uintptr_t)kgem->exec;
 			execbuf.buffer_count = kgem->nexec;
 			execbuf.batch_start_offset = 0;
@@ -1001,7 +1048,7 @@ void _kgem_submit(struct kgem *kgem)
 					       found ? found->size : -1,
 					       found ? found->tiling : -1,
 					       (int)(kgem->exec[i].flags & EXEC_OBJECT_NEEDS_FENCE),
-					       found ? found->deleted : -1);
+					       found ? found->purged : -1);
 				}
 				for (i = 0; i < kgem->nreloc; i++) {
 					ErrorF("reloc[%d] = pos:%d, target:%d, delta:%d, read:%x, write:%x, offset:%x\n",
@@ -1022,6 +1069,7 @@ void _kgem_submit(struct kgem *kgem)
 				struct drm_i915_gem_set_domain set_domain;
 				int ret;
 
+				VG_CLEAR(set_domain);
 				set_domain.handle = handle;
 				set_domain.read_domains = I915_GEM_DOMAIN_GTT;
 				set_domain.write_domain = I915_GEM_DOMAIN_GTT;
@@ -1041,7 +1089,7 @@ void _kgem_submit(struct kgem *kgem)
 		kgem_cleanup(kgem);
 
 	kgem_reset(kgem);
-	kgem->busy = 1;
+	kgem->flush_now = kgem->busy = 1;
 }
 
 void kgem_throttle(struct kgem *kgem)
@@ -1096,7 +1144,7 @@ bool kgem_expire_cache(struct kgem *kgem)
 	for (i = 0; i < ARRAY_SIZE(kgem->inactive); i++) {
 		idle &= list_is_empty(&kgem->inactive[i]);
 		list_for_each_entry(bo, &kgem->inactive[i], list) {
-			assert(bo->deleted);
+			assert(bo->purged);
 			if (bo->delta) {
 				expire = now - 5;
 				break;
@@ -1131,9 +1179,7 @@ bool kgem_expire_cache(struct kgem *kgem)
 			count++;
 			size += bo->size;
 
-			gem_close(kgem->fd, bo->handle);
-			list_del(&bo->list);
-			free(bo);
+			kgem_bo_free(kgem, bo);
 		}
 	}
 
@@ -1149,7 +1195,6 @@ bool kgem_expire_cache(struct kgem *kgem)
 
 void kgem_cleanup_cache(struct kgem *kgem)
 {
-	struct kgem_bo *bo;
 	unsigned int i;
 
 	/* sync to the most recent request */
@@ -1161,6 +1206,7 @@ void kgem_cleanup_cache(struct kgem *kgem)
 				      struct kgem_request,
 				      list);
 
+		VG_CLEAR(set_domain);
 		set_domain.handle = rq->bo->handle;
 		set_domain.read_domains = I915_GEM_DOMAIN_GTT;
 		set_domain.write_domain = I915_GEM_DOMAIN_GTT;
@@ -1172,14 +1218,10 @@ void kgem_cleanup_cache(struct kgem *kgem)
 	kgem_expire_partial(kgem);
 
 	for (i = 0; i < ARRAY_SIZE(kgem->inactive); i++) {
-		while (!list_is_empty(&kgem->inactive[i])) {
-			bo = list_last_entry(&kgem->inactive[i],
-					     struct kgem_bo, list);
-
-			gem_close(kgem->fd, bo->handle);
-			list_del(&bo->list);
-			free(bo);
-		}
+		while (!list_is_empty(&kgem->inactive[i]))
+			kgem_bo_free(kgem,
+				     list_last_entry(&kgem->inactive[i],
+						     struct kgem_bo, list));
 	}
 
 	kgem->need_purge = false;
@@ -1200,14 +1242,14 @@ search_linear_cache(struct kgem *kgem, unsigned int size, bool use_active)
 		if (use_active && bo->tiling != I915_TILING_NONE)
 			continue;
 
-		if (bo->deleted) {
+		if (bo->purged) {
 			if (!gem_madvise(kgem->fd, bo->handle,
 					 I915_MADV_WILLNEED)) {
 				kgem->need_purge |= bo->gpu;
 				continue;
 			}
 
-			bo->deleted = 0;
+			bo->purged = false;
 		}
 
 		if (I915_TILING_NONE != bo->tiling &&
@@ -1227,6 +1269,7 @@ search_linear_cache(struct kgem *kgem, unsigned int size, bool use_active)
 		     use_active ? "active" : "inactive"));
 		assert(bo->refcnt == 0);
 		assert(bo->reusable);
+		assert(use_active || bo->purgeable);
 		assert(use_active || bo->gpu == 0);
 		//assert(use_active || !kgem_busy(kgem, bo->handle));
 		return bo;
@@ -1242,7 +1285,7 @@ struct kgem_bo *kgem_create_for_name(struct kgem *kgem, uint32_t name)
 
 	DBG(("%s(name=%d)\n", __FUNCTION__, name));
 
-	memset(&open_arg, 0, sizeof(open_arg));
+	VG_CLEAR(open_arg);
 	open_arg.name = name;
 	if (drmIoctl(kgem->fd, DRM_IOCTL_GEM_OPEN, &open_arg))
 		return NULL;
@@ -1255,6 +1298,7 @@ struct kgem_bo *kgem_create_for_name(struct kgem *kgem, uint32_t name)
 	}
 
 	bo->reusable = false;
+	bo->purgeable = false;
 	return bo;
 }
 
@@ -1271,10 +1315,11 @@ struct kgem_bo *kgem_create_linear(struct kgem *kgem, int size)
 		return kgem_bo_reference(bo);
 
 	if (!list_is_empty(&kgem->requests)) {
-		kgem_retire(kgem);
-		bo = search_linear_cache(kgem, size, false);
-		if (bo)
-			return kgem_bo_reference(bo);
+		if (kgem_retire(kgem)) {
+			bo = search_linear_cache(kgem, size, false);
+			if (bo)
+				return kgem_bo_reference(bo);
+		}
 	}
 
 	handle = gem_create(kgem->fd, size);
@@ -1374,9 +1419,12 @@ static bool _kgem_can_create_2d(struct kgem *kgem,
 	if (tiling < 0)
 		tiling = -tiling;
 
-	size = kgem_surface_size(kgem, width, height, bpp, tiling, &pitch);
-	if (size == 0 || size > kgem->max_object_size)
-		size = kgem_surface_size(kgem, width, height, bpp, I915_TILING_NONE, &pitch);
+	size = kgem_surface_size(kgem, false,
+				 width, height, bpp, tiling, &pitch);
+	if (size == 0 || size >= kgem->max_object_size)
+		size = kgem_surface_size(kgem, false,
+					 width, height, bpp,
+					 I915_TILING_NONE, &pitch);
 	return size > 0 && size <= kgem->max_object_size;
 }
 
@@ -1434,8 +1482,9 @@ struct kgem_bo *kgem_create_2d(struct kgem *kgem,
 	DBG(("%s(%dx%d, bpp=%d, tiling=%d, exact=%d, inactive=%d)\n", __FUNCTION__,
 	     width, height, bpp, tiling, !!exact, !!(flags & CREATE_INACTIVE)));
 
-	assert(_kgem_can_create_2d(kgem, width, height, bpp, tiling));
-	size = kgem_surface_size(kgem, width, height, bpp, tiling, &pitch);
+	assert(_kgem_can_create_2d(kgem, width, height, bpp, exact ? -tiling : tiling));
+	size = kgem_surface_size(kgem, kgem->has_relaxed_fencing,
+				 width, height, bpp, tiling, &pitch);
 	assert(size && size <= kgem->max_object_size);
 	if (flags & CREATE_INACTIVE)
 		goto skip_active_search;
@@ -1471,18 +1520,16 @@ search_active: /* Best active match first */
 			if (bo->rq == NULL)
 				list_del(&bo->request);
 
-			if (bo->deleted) {
+			if (bo->purged) {
 				if (!gem_madvise(kgem->fd, bo->handle,
 						 I915_MADV_WILLNEED)) {
 					kgem->need_purge |= bo->gpu;
-					gem_close(kgem->fd, bo->handle);
-					list_del(&bo->request);
-					free(bo);
+					kgem_bo_free(kgem, bo);
 					bo = NULL;
 					goto search_active;
 				}
 
-				bo->deleted = 0;
+				bo->purged = false;
 			}
 
 			bo->unique_id = kgem_get_unique_id(kgem);
@@ -1519,14 +1566,14 @@ skip_active_search:
 		list_del(&bo->list);
 		assert(list_is_empty(&bo->request));
 
-		if (bo->deleted) {
+		if (bo->purged) {
 			if (!gem_madvise(kgem->fd, bo->handle,
 					 I915_MADV_WILLNEED)) {
 				kgem->need_purge |= bo->gpu;
 				goto next_bo;
 			}
 
-			bo->deleted = 0;
+			bo->purged = false;
 		}
 
 		bo->delta = 0;
@@ -1536,22 +1583,22 @@ skip_active_search:
 		     bo->pitch, bo->tiling, bo->handle, bo->unique_id));
 		assert(bo->refcnt == 0);
 		assert(bo->reusable);
+		assert(bo->purgeable);
 		assert((flags & CREATE_INACTIVE) == 0 || bo->gpu == 0);
 		assert((flags & CREATE_INACTIVE) == 0 ||
 		       !kgem_busy(kgem, bo->handle));
 		return kgem_bo_reference(bo);
 
 next_bo:
-		gem_close(kgem->fd, bo->handle);
-		list_del(&bo->request);
-		free(bo);
+		kgem_bo_free(kgem, bo);
 		continue;
 	}
 
 	if (flags & CREATE_INACTIVE && !list_is_empty(&kgem->requests)) {
-		kgem_retire(kgem);
-		flags &= ~CREATE_INACTIVE;
-		goto skip_active_search;
+		if (kgem_retire(kgem)) {
+			flags &= ~CREATE_INACTIVE;
+			goto skip_active_search;
+		}
 	}
 
 	handle = gem_create(kgem->fd, size);
@@ -1580,6 +1627,8 @@ void _kgem_bo_destroy(struct kgem *kgem, struct kgem_bo *bo)
 {
 	if (bo->proxy) {
 		kgem_bo_unref(kgem, bo->proxy);
+
+		assert(bo->binding.next == NULL);
 		list_del(&bo->request);
 		free(bo);
 		return;
@@ -1687,7 +1736,7 @@ uint32_t kgem_add_reloc(struct kgem *kgem,
 	assert(index < ARRAY_SIZE(kgem->reloc));
 	kgem->reloc[index].offset = pos * sizeof(kgem->batch[0]);
 	if (bo) {
-		assert(!bo->deleted);
+		assert(!bo->purged);
 
 		delta += bo->delta;
 		if (bo->proxy) {
@@ -1697,7 +1746,7 @@ uint32_t kgem_add_reloc(struct kgem *kgem,
 			bo = bo->proxy;
 		}
 
-		assert(!bo->deleted);
+		assert(!bo->purged);
 
 		if (bo->exec == NULL) {
 			_kgem_add_bo(kgem, bo);
@@ -1756,7 +1805,7 @@ uint32_t kgem_bo_flink(struct kgem *kgem, struct kgem_bo *bo)
 	struct drm_gem_flink flink;
 	int ret;
 
-	memset(&flink, 0, sizeof(flink));
+	VG_CLEAR(flink);
 	flink.handle = bo->handle;
 	ret = drmIoctl(kgem->fd, DRM_IOCTL_GEM_FLINK, &flink);
 	if (ret)
@@ -1767,6 +1816,7 @@ uint32_t kgem_bo_flink(struct kgem *kgem, struct kgem_bo *bo)
 	 * on the buffer, and *presuming* they do not pass it on to a third
 	 * party, we track the lifetime accurately.
 	 */
+	bo->purgeable = false;
 	return flink.name;
 }
 
@@ -1775,6 +1825,7 @@ static uint32_t gem_vmap(int fd, void *ptr, int size, int read_only)
 {
 	struct drm_i915_gem_vmap vmap;
 
+	VG_CLEAR(vmap);
 	vmap.user_ptr = (uintptr_t)ptr;
 	vmap.user_size = size;
 	vmap.flags = 0;
@@ -1807,7 +1858,7 @@ struct kgem_bo *kgem_create_map(struct kgem *kgem,
 		return NULL;
 	}
 
-	bo->reusable = false;
+	bo->purgeable = bo->reusable = false;
 	bo->sync = true;
 	DBG(("%s(ptr=%p, size=%d, read_only=%d) => handle=%d\n",
 	     __FUNCTION__, ptr, size, read_only, handle));
@@ -1843,19 +1894,44 @@ void kgem_bo_sync(struct kgem *kgem, struct kgem_bo *bo, bool for_write)
 	if (for_write ? bo->cpu_write : bo->cpu_read)
 		return;
 
+	VG_CLEAR(set_domain);
 	set_domain.handle = bo->handle;
 	set_domain.read_domains = I915_GEM_DOMAIN_CPU;
 	set_domain.write_domain = for_write ? I915_GEM_DOMAIN_CPU : 0;
 
 	drmIoctl(kgem->fd, DRM_IOCTL_I915_GEM_SET_DOMAIN, &set_domain);
 	bo->needs_flush = false;
-	if (bo->gpu)
+	if (bo->gpu) {
+		kgem->sync = false;
 		kgem_retire(kgem);
+	}
 	bo->cpu_read = true;
 	if (for_write)
 		bo->cpu_write = true;
 }
 
+void kgem_sync(struct kgem *kgem)
+{
+	if (!list_is_empty(&kgem->requests)) {
+		struct drm_i915_gem_set_domain set_domain;
+		struct kgem_request *rq;
+
+		rq = list_first_entry(&kgem->requests,
+				      struct kgem_request,
+				      list);
+
+		VG_CLEAR(set_domain);
+		set_domain.handle = rq->bo->handle;
+		set_domain.read_domains = I915_GEM_DOMAIN_GTT;
+		set_domain.write_domain = I915_GEM_DOMAIN_GTT;
+
+		drmIoctl(kgem->fd, DRM_IOCTL_I915_GEM_SET_DOMAIN, &set_domain);
+		kgem_retire(kgem);
+	}
+
+	kgem->sync = false;
+}
+
 void kgem_clear_dirty(struct kgem *kgem)
 {
 	struct kgem_request *rq = kgem->next_request;
@@ -1911,7 +1987,7 @@ struct kgem_bo *kgem_create_proxy(struct kgem_bo *target,
 	if (bo == NULL)
 		return NULL;
 
-	bo->reusable = false;
+	bo->purgeable = bo->reusable = false;
 	bo->proxy = kgem_bo_reference(target);
 	bo->delta = offset;
 	return bo;
@@ -1948,20 +2024,28 @@ struct kgem_bo *kgem_create_buffer(struct kgem *kgem,
 {
 	struct kgem_partial_bo *bo;
 	bool write = !!(flags & KGEM_BUFFER_WRITE);
-	int offset, alloc;
+	unsigned offset, alloc;
 	uint32_t handle;
 
 	DBG(("%s: size=%d, flags=%x [write=%d, last=%d]\n",
 	     __FUNCTION__, size, flags, write, flags & KGEM_BUFFER_LAST));
+	assert(size);
 
 	list_for_each_entry(bo, &kgem->partial, base.list) {
 		if (flags == KGEM_BUFFER_LAST && bo->write) {
 			/* We can reuse any write buffer which we can fit */
-			if (size < bo->alloc) {
-				DBG(("%s: reusing write buffer for read of %d bytes? used=%d, total=%d\n",
-				     __FUNCTION__, size, bo->used, bo->alloc));
-				offset = 0;
-				goto done;
+			if (size <= bo->alloc) {
+				if (bo->base.refcnt == 1 && bo->base.exec) {
+					DBG(("%s: reusing write buffer for read of %d bytes? used=%d, total=%d\n",
+					     __FUNCTION__, size, bo->used, bo->alloc));
+					offset = 0;
+					goto done;
+				} else if (bo->used + size <= bo->alloc) {
+					DBG(("%s: reusing unfinished write buffer for read of %d bytes? used=%d, total=%d\n",
+					     __FUNCTION__, size, bo->used, bo->alloc));
+					offset = bo->used;
+					goto done;
+				}
 			}
 		}
 
@@ -1971,11 +2055,14 @@ struct kgem_bo *kgem_create_buffer(struct kgem *kgem,
 			continue;
 		}
 
-		if (bo->base.refcnt == 1 && bo->base.exec == NULL)
+		if (bo->base.refcnt == 1 && bo->base.exec == NULL) {
+			DBG(("%s: discarding unfinished buffer? used=%d, total=%d\n",
+			     __FUNCTION__, bo->used, bo->alloc));
 			/* no users, so reset */
 			bo->used = 0;
+		}
 
-		if (bo->used + size < bo->alloc) {
+		if (bo->used + size <= bo->alloc) {
 			DBG(("%s: reusing partial buffer? used=%d + size=%d, total=%d\n",
 			     __FUNCTION__, bo->used, size, bo->alloc));
 			offset = bo->used;
@@ -2023,13 +2110,13 @@ struct kgem_bo *kgem_create_buffer(struct kgem *kgem,
 				return NULL;
 			}
 		}
-		bo->need_io = true;
+		bo->need_io = write;
 	} else {
 		__kgem_bo_init(&bo->base, handle, alloc);
-		bo->base.sync = true;
+		bo->base.vmap = true;
 		bo->need_io = 0;
 	}
-	bo->base.reusable = false;
+	bo->base.purgeable = bo->base.reusable = false;
 
 	bo->alloc = alloc;
 	bo->used = size;
@@ -2164,15 +2251,53 @@ void kgem_buffer_read_sync(struct kgem *kgem, struct kgem_bo *_bo)
 
 	bo = (struct kgem_partial_bo *)_bo;
 
-	DBG(("%s(offset=%d, length=%d, sync=%d)\n", __FUNCTION__,
-	     offset, length, bo->base.sync));
+	DBG(("%s(offset=%d, length=%d, vmap=%d)\n", __FUNCTION__,
+	     offset, length, bo->base.vmap));
 
-	if (!bo->base.sync) {
-		gem_read(kgem->fd, bo->base.handle,
-			 (char *)(bo+1)+offset, length);
+	if (!bo->base.vmap) {
+		gem_read(kgem->fd,
+			 bo->base.handle, (char *)(bo+1)+offset,
+			 offset, length);
 		bo->base.needs_flush = false;
 		if (bo->base.gpu)
 			kgem_retire(kgem);
 	} else
 		kgem_bo_sync(kgem, &bo->base, false);
 }
+
+uint32_t kgem_bo_get_binding(struct kgem_bo *bo, uint32_t format)
+{
+	struct kgem_bo_binding *b;
+
+	for (b = &bo->binding; b && b->offset; b = b->next)
+		if (format == b->format)
+			return b->offset;
+
+	return 0;
+}
+
+void kgem_bo_set_binding(struct kgem_bo *bo, uint32_t format, uint16_t offset)
+{
+	struct kgem_bo_binding *b;
+
+	for (b = &bo->binding; b; b = b->next) {
+		if (b->offset)
+			continue;
+
+		b->offset = offset;
+		b->format = format;
+
+		if (b->next)
+			b->next->offset = 0;
+
+		return;
+	}
+
+	b = malloc(sizeof(*b));
+	if (b) {
+		b->next = bo->binding.next;
+		b->format = format;
+		b->offset = offset;
+		bo->binding.next = b;
+	}
+}
