@@ -51,7 +51,12 @@ struct kgem_bo {
 	struct kgem_request *rq;
 	struct drm_i915_gem_exec_object2 *exec;
 
-	uint16_t src_bound, dst_bound;
+	struct kgem_bo_binding {
+		struct kgem_bo_binding *next;
+		uint32_t format;
+		uint16_t offset;
+	} binding;
+
 	uint32_t unique_id;
 	uint32_t refcnt;
 	uint32_t handle;
@@ -59,7 +64,7 @@ struct kgem_bo {
 	uint32_t size;
 	uint32_t delta;
 
-	uint32_t pitch : 16;
+	uint32_t pitch : 18; /* max 128k */
 	uint32_t tiling : 2;
 	uint32_t reusable : 1;
 	uint32_t dirty : 1;
@@ -67,9 +72,11 @@ struct kgem_bo {
 	uint32_t needs_flush : 1;
 	uint32_t cpu_read : 1;
 	uint32_t cpu_write : 1;
+	uint32_t vmap : 1;
 	uint32_t flush : 1;
 	uint32_t sync : 1;
-	uint32_t deleted : 1;
+	uint32_t purgeable : 1;
+	uint32_t purged : 1;
 };
 
 struct kgem_request {
@@ -105,8 +112,11 @@ struct kgem {
 	uint16_t nfence;
 
 	uint32_t flush:1;
+	uint32_t sync:1;
 	uint32_t need_expire:1;
 	uint32_t need_purge:1;
+	uint32_t need_retire:1;
+	uint32_t flush_now:1;
 	uint32_t busy:1;
 
 	uint32_t has_vmap :1;
@@ -171,7 +181,10 @@ struct kgem_bo *kgem_create_2d(struct kgem *kgem,
 			       int tiling,
 			       uint32_t flags);
 
-void kgem_retire(struct kgem *kgem);
+uint32_t kgem_bo_get_binding(struct kgem_bo *bo, uint32_t format);
+void kgem_bo_set_binding(struct kgem_bo *bo, uint32_t format, uint16_t offset);
+
+bool kgem_retire(struct kgem *kgem);
 
 void _kgem_submit(struct kgem *kgem);
 static inline void kgem_submit(struct kgem *kgem)
@@ -271,8 +284,8 @@ static inline void kgem_advance_batch(struct kgem *kgem, int num_dwords)
 	kgem->nbatch += num_dwords;
 }
 
-bool kgem_check_bo(struct kgem *kgem, ...) __attribute__((sentinel(NULL)));
-bool kgem_check_bo_fenced(struct kgem *kgem, ...) __attribute__((sentinel(NULL)));
+bool kgem_check_bo(struct kgem *kgem, ...) __attribute__((sentinel(0)));
+bool kgem_check_bo_fenced(struct kgem *kgem, ...) __attribute__((sentinel(0)));
 
 void _kgem_add_bo(struct kgem *kgem, struct kgem_bo *bo);
 static inline void kgem_add_bo(struct kgem *kgem, struct kgem_bo *bo)
@@ -324,6 +337,7 @@ static inline void kgem_bo_mark_dirty(struct kgem_bo *bo)
 }
 
 void kgem_bo_sync(struct kgem *kgem, struct kgem_bo *bo, bool for_write);
+void kgem_sync(struct kgem *kgem);
 
 #define KGEM_BUFFER_WRITE	0x1
 #define KGEM_BUFFER_LAST	0x2
