@@ -44,8 +44,6 @@
 #if DEBUG_BLT
 #undef DBG
 #define DBG(x) ErrorF x
-#else
-#define NDEBUG 1
 #endif
 
 #define NO_BLT_COMPOSITE 0
@@ -139,11 +137,16 @@ static bool sna_blt_fill_init(struct sna *sna,
 		blt->cmd |= 1 << 11;
 		pitch >>= 2;
 	}
-	if (pitch > MAXSHORT)
-		return FALSE;
+	assert(pitch < MAXSHORT);
 
-	if (pixel == 0)
-		alu = GXclear;
+	if (alu == GXclear)
+		pixel = 0;
+	else if (alu == GXcopy) {
+		if (pixel == 0)
+			alu = GXclear;
+		else if (pixel == -1)
+			alu = GXset;
+	}
 
 	blt->br13 = 1<<31 | (fill_ROP[alu] << 16) | pitch;
 	switch (bpp) {
@@ -164,7 +167,8 @@ static bool sna_blt_fill_init(struct sna *sna,
 	}
 
 	if (sna->blt_state.fill_bo != bo->handle ||
-	    sna->blt_state.fill_pixel != pixel)
+	    sna->blt_state.fill_pixel != pixel ||
+	    sna->blt_state.fill_alu != alu)
 	{
 		uint32_t *b;
 
@@ -193,6 +197,7 @@ static bool sna_blt_fill_init(struct sna *sna,
 
 		sna->blt_state.fill_bo = bo->handle;
 		sna->blt_state.fill_pixel = pixel;
+		sna->blt_state.fill_alu = alu;
 	}
 
 	return TRUE;
@@ -274,18 +279,16 @@ static Bool sna_blt_copy_init(struct sna *sna,
 		blt->cmd |= BLT_SRC_TILED;
 		blt->pitch[0] >>= 2;
 	}
-	if (blt->pitch[0] > MAXSHORT)
-		return FALSE;
+	assert(blt->pitch[0] < MAXSHORT);
 
 	blt->pitch[1] = dst->pitch;
 	if (kgem->gen >= 40 && dst->tiling) {
 		blt->cmd |= BLT_DST_TILED;
 		blt->pitch[1] >>= 2;
 	}
-	if (blt->pitch[1] > MAXSHORT)
-		return FALSE;
+	assert(blt->pitch[1] < MAXSHORT);
 
-	blt->overwrites = alu == GXcopy || alu == GXclear;
+	blt->overwrites = alu == GXcopy || alu == GXclear || alu == GXset;
 	blt->br13 = (copy_ROP[alu] << 16) | blt->pitch[1];
 	switch (bpp) {
 	default: assert(0);
@@ -328,7 +331,7 @@ static void sna_blt_copy_one(struct sna *sna,
 	/* Compare against a previous fill */
 	if (kgem->nbatch >= 6 &&
 	    blt->overwrites &&
-	    kgem->batch[kgem->nbatch-6] == ((blt->cmd & ~XY_SRC_COPY_BLT_CMD) | XY_COLOR_BLT_CMD) &&
+	    kgem->batch[kgem->nbatch-6] == ((blt->cmd & ~XY_SRC_COPY_BLT_CMD) | XY_COLOR_BLT) &&
 	    kgem->batch[kgem->nbatch-4] == ((uint32_t)dst_y << 16 | (uint16_t)dst_x) &&
 	    kgem->batch[kgem->nbatch-3] == ((uint32_t)(dst_y+height) << 16 | (uint16_t)(dst_x+width)) &&
 	    kgem->reloc[kgem->nreloc-1].target_handle == blt->bo[1]->handle) {
@@ -630,7 +633,7 @@ pixel_is_white(uint32_t pixel, uint32_t format)
 	case PICT_TYPE_ARGB:
 	case PICT_TYPE_ABGR:
 	case PICT_TYPE_BGRA:
-		return pixel == ((1 << PICT_FORMAT_BPP(format)) - 1);
+		return pixel == ((1U << PICT_FORMAT_BPP(format)) - 1);
 	default:
 		return FALSE;
 	}
@@ -1072,9 +1075,10 @@ static void blt_put_composite_boxes(struct sna *sna,
 	PixmapPtr src = op->u.blt.src_pixmap;
 	struct sna_pixmap *dst_priv = sna_pixmap(op->dst.pixmap);
 
-	DBG(("%s: src=(%d, %d), dst=(%d, %d) x %d\n", __FUNCTION__,
+	DBG(("%s: src=(%d, %d), dst=(%d, %d), [(%d, %d), (%d, %d) x %d]\n", __FUNCTION__,
 	     op->u.blt.sx, op->u.blt.sy,
-	     op->dst.x, op->dst.y, n));
+	     op->dst.x, op->dst.y,
+	     box->x1, box->y1, box->x2, box->y2, n));
 
 	if (n == 1 && !dst_priv->pinned &&
 	    box->x2 - box->x1 == op->dst.width &&
@@ -1118,13 +1122,8 @@ prepare_blt_put(struct sna *sna,
 	if (priv) {
 		if (!priv->gpu_only) {
 			src_bo = priv->cpu_bo;
-			if (!src_bo) {
-				src_bo = kgem_create_map(&sna->kgem,
-							 src->devPrivate.ptr,
-							 pixmap_size(src),
-							 1);
-				priv->cpu_bo = src_bo;
-			}
+			if (!src_bo)
+				src_bo = pixmap_vmap(&sna->kgem, src);
 		}
 	} else {
 		src_bo = kgem_create_map(&sna->kgem,
@@ -1175,6 +1174,7 @@ has_gpu_area(PixmapPtr pixmap, int x, int y, int w, int h)
 	area.y1 = y;
 	area.x2 = x + w;
 	area.y2 = y + h;
+
 	return sna_damage_contains_box(priv->cpu_damage,
 				       &area) == PIXMAN_REGION_OUT;
 }
@@ -1271,8 +1271,7 @@ sna_blt_composite(struct sna *sna,
 	get_drawable_deltas(dst->pDrawable, tmp->dst.pixmap,
 			    &tmp->dst.x, &tmp->dst.y);
 	tmp->dst.bo = priv->gpu_bo;
-	if (!priv->gpu_only &&
-	    !sna_damage_is_all(&priv->gpu_damage,
+	if (!sna_damage_is_all(&priv->gpu_damage,
 			       tmp->dst.width, tmp->dst.height))
 		tmp->damage = &priv->gpu_damage;
 	if (width && height)
@@ -1309,11 +1308,9 @@ sna_blt_composite(struct sna *sna,
 		return FALSE;
 	}
 
-	if (!sna_blt_compare_depth(src->pDrawable, dst->pDrawable)) {
-		DBG(("%s: mismatching depth src=%d/%d, dst=%d/%d\n",
-		     __FUNCTION__,
-		     src->pDrawable->depth, src->pDrawable->bitsPerPixel,
-		     dst->pDrawable->depth, dst->pDrawable->bitsPerPixel));
+	if (src->filter == PictFilterConvolution) {
+		DBG(("%s: convolutions filters not handled\n",
+		     __FUNCTION__));
 		return FALSE;
 	}
 
@@ -1326,12 +1323,6 @@ sna_blt_composite(struct sna *sna,
 		return FALSE;
 	}
 
-	if (src->filter == PictFilterConvolution) {
-		DBG(("%s: convolutions filters not handled\n",
-		     __FUNCTION__));
-		return FALSE;
-	}
-
 	if (!(dst->format == src_format ||
 	      dst->format == PICT_FORMAT(PICT_FORMAT_BPP(src_format),
 					 PICT_FORMAT_TYPE(src_format),
@@ -1504,8 +1495,7 @@ static Bool sna_blt_fill_box(struct sna *sna, uint8_t alu,
 			     const BoxRec *box)
 {
 	struct kgem *kgem = &sna->kgem;
-	uint32_t br13, cmd;
-	uint32_t *b;
+	uint32_t br13, cmd, *b;
 
 	DBG(("%s: box=((%d, %d), (%d, %d))\n", __FUNCTION__,
 	     box->x1, box->y1, box->x2, box->y2));
@@ -1513,38 +1503,37 @@ static Bool sna_blt_fill_box(struct sna *sna, uint8_t alu,
 	assert(box->x1 >= 0);
 	assert(box->y1 >= 0);
 
-	cmd = XY_COLOR_BLT_CMD;
-	if (bpp == 32)
-		cmd |= BLT_WRITE_ALPHA | BLT_WRITE_RGB;
-
+	cmd = XY_COLOR_BLT;
 	br13 = bo->pitch;
 	if (kgem->gen >= 40 && bo->tiling) {
 		cmd |= BLT_DST_TILED;
 		br13 >>= 2;
 	}
-	if (br13 > MAXSHORT) {
-		DBG(("%s: fallback -- pitch is too large %d [%d]\n",
-		     __FUNCTION__, bo->pitch, br13));
-		return FALSE;
-	}
+	assert(br13 < MAXSHORT);
 
-	if (color == 0)
-		alu = GXclear;
+	if (alu == GXclear)
+		color = 0;
+	else if (alu == GXcopy) {
+		if (color == 0)
+			alu = GXclear;
+		else if (color == -1)
+			alu = GXset;
+	}
 
 	br13 |= fill_ROP[alu] << 16;
 	switch (bpp) {
 	default: assert(0);
-	case 32: br13 |= 1 << 25; /* RGB8888 */
+	case 32: cmd |= BLT_WRITE_ALPHA | BLT_WRITE_RGB;
+		 br13 |= 1 << 25; /* RGB8888 */
 	case 16: br13 |= 1 << 24; /* RGB565 */
 	case 8: break;
 	}
 
 	/* All too frequently one blt completely overwrites the previous */
 	if (kgem->nbatch >= 6 &&
-	    (alu == GXcopy || alu == GXclear) &&
+	    (alu == GXcopy || alu == GXclear || alu == GXset) &&
 	    kgem->batch[kgem->nbatch-6] == cmd &&
-	    kgem->batch[kgem->nbatch-4] == ((uint32_t)box[0].y1 << 16 | (uint16_t)box[0].x1) &&
-	    kgem->batch[kgem->nbatch-3] == ((uint32_t)box[0].y2 << 16 | (uint16_t)box[0].x2) &&
+	    *(uint64_t *)&kgem->batch[kgem->nbatch-4] == *(uint64_t *)box &&
 	    kgem->reloc[kgem->nreloc-1].target_handle == bo->handle) {
 		DBG(("%s: replacing last fill\n", __FUNCTION__));
 		kgem->batch[kgem->nbatch-5] = br13;
@@ -1561,12 +1550,10 @@ static Bool sna_blt_fill_box(struct sna *sna, uint8_t alu,
 	}
 
 	b = kgem->batch + kgem->nbatch;
-
 	b[0] = cmd;
 	b[1] = br13;
 	*(uint64_t *)(b+2) = *(uint64_t *)box;
-	b[4] = kgem_add_reloc(kgem, kgem->nbatch + 4,
-			      bo,
+	b[4] = kgem_add_reloc(kgem, kgem->nbatch + 4, bo,
 			      I915_GEM_DOMAIN_RENDER << 16 |
 			      I915_GEM_DOMAIN_RENDER |
 			      KGEM_RELOC_FENCED,
@@ -1607,11 +1594,16 @@ Bool sna_blt_fill_boxes(struct sna *sna, uint8_t alu,
 		cmd |= 1 << 11;
 		br13 >>= 2;
 	}
-	if (br13 > MAXSHORT)
-		return FALSE;
+	assert(br13 < MAXSHORT);
 
-	if (pixel == 0)
-		alu = GXclear;
+	if (alu == GXclear)
+		pixel = 0;
+	else if (alu == GXcopy) {
+		if (pixel == 0)
+			alu = GXclear;
+		else if (pixel == -1)
+			alu = GXset;
+	}
 
 	br13 |= 1<<31 | fill_ROP[alu] << 16;
 	switch (bpp) {
@@ -1629,7 +1621,8 @@ Bool sna_blt_fill_boxes(struct sna *sna, uint8_t alu,
 	}
 
 	if (sna->blt_state.fill_bo != bo->handle ||
-	    sna->blt_state.fill_pixel != pixel)
+	    sna->blt_state.fill_pixel != pixel ||
+	    sna->blt_state.fill_alu != alu)
 	{
 		uint32_t *b;
 
@@ -1658,6 +1651,7 @@ Bool sna_blt_fill_boxes(struct sna *sna, uint8_t alu,
 
 		sna->blt_state.fill_bo = bo->handle;
 		sna->blt_state.fill_pixel = pixel;
+		sna->blt_state.fill_alu = alu;
 	}
 
 	do {
@@ -1724,7 +1718,7 @@ Bool sna_blt_copy_boxes(struct sna *sna, uint8_t alu,
 			int bpp, const BoxRec *box, int nbox)
 {
 	struct kgem *kgem = &sna->kgem;
-	int src_pitch, br13, cmd;
+	unsigned src_pitch, br13, cmd;
 
 #if DEBUG_NO_BLT || NO_BLT_COPY_BOXES
 	return FALSE;
@@ -1747,16 +1741,14 @@ Bool sna_blt_copy_boxes(struct sna *sna, uint8_t alu,
 		cmd |= BLT_SRC_TILED;
 		src_pitch >>= 2;
 	}
-	if (src_pitch > MAXSHORT)
-		return FALSE;
+	assert(src_pitch < MAXSHORT);
 
 	br13 = dst_bo->pitch;
 	if (kgem->gen >= 40 && dst_bo->tiling) {
 		cmd |= BLT_DST_TILED;
 		br13 >>= 2;
 	}
-	if (br13 > MAXSHORT)
-		return FALSE;
+	assert(br13 < MAXSHORT);
 
 	br13 |= copy_ROP[alu] << 16;
 	switch (bpp) {
@@ -1768,9 +1760,9 @@ Bool sna_blt_copy_boxes(struct sna *sna, uint8_t alu,
 
 	/* Compare first box against a previous fill */
 	if (kgem->nbatch >= 6 &&
-	    (alu == GXcopy || alu == GXclear) &&
+	    (alu == GXcopy || alu == GXclear || alu == GXset) &&
 	    kgem->reloc[kgem->nreloc-1].target_handle == dst_bo->handle &&
-	    kgem->batch[kgem->nbatch-6] == ((cmd & ~XY_SRC_COPY_BLT_CMD) | XY_COLOR_BLT_CMD) &&
+	    kgem->batch[kgem->nbatch-6] == ((cmd & ~XY_SRC_COPY_BLT_CMD) | XY_COLOR_BLT) &&
 	    kgem->batch[kgem->nbatch-4] == ((uint32_t)(box->y1 + dst_dy) << 16 | (uint16_t)(box->x1 + dst_dx)) &&
 	    kgem->batch[kgem->nbatch-3] == ((uint32_t)(box->y2 + dst_dy) << 16 | (uint16_t)(box->x2 + dst_dx))) {
 		DBG(("%s: deleting last fill\n", __FUNCTION__));
